#include "uxn.h"
#include "dat.h"
#include "fns.h"
#include "dev/screen.h"

static Uint8 blending[5][16] = {
	{0, 0, 0, 0, 1, 0, 1, 1, 2, 2, 0, 2, 3, 3, 3, 0},
	{0, 1, 2, 3, 0, 1, 2, 3, 0, 1, 2, 3, 0, 1, 2, 3},
	{1, 2, 3, 1, 1, 2, 3, 1, 1, 2, 3, 1, 1, 2, 3, 1},
	{2, 3, 1, 2, 2, 3, 1, 2, 2, 3, 1, 2, 2, 3, 1, 2},
	{1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 0, 1, 1, 1, 1, 0}
};

void
screen_init(Screen *scr)
{
	Uint32 i;
	Uint32 size;

	scr->pixels = (Uint16*) PHYSFB;
	scr->width  = PHYSFBWIDTH;
	scr->height = PHYSFBHEIGHT;

	size = scr->width * scr->height;
	for(i = 0; i < size; i++) {
		scr->fg.pixels[i] = 0;
		scr->bg.pixels[i] = 0;
	}

	scr->fg.changed = 1;
	scr->bg.changed = 1;
}

void
screen_redraw(Screen *scr)
{
	Uint32 i;
	Uint32 size;
	Uint16 palette[16];

	for(i = 0; i < 16; i++)
		palette[i] = scr->palette[(i >> 2) ? (i >> 2) : (i & 3)];

	size = scr->width * scr->height;
	for(i = 0; i < size; i++)
		scr->pixels[i] = palette[scr->fg.pixels[i] << 2 | scr->bg.pixels[i]];

	scr->fg.changed = 0;
	scr->bg.changed = 0;
}

static void
screen_write(Screen *scr, Layer *layer, Uint16 x, Uint16 y, Uint8 color)
{
	if(x < scr->width && y < scr->height) {
		Uint32 i = x + y * scr->width;

		if(color != layer->pixels[i]) {
			layer->pixels[i] = color;
			layer->changed = 1;
		}
	}
}

static void
screen_blit(Screen *scr, Layer *layer, Uint16 x, Uint16 y, Uint8 *sprite, Uint8 color, Uint8 flags)
{
	int v, h;
	int opaque;
	Uint8 flipx, flipy, twobpp;

	opaque = blending[4][color];
	flipx  = flags & 0x10;
	flipy  = flags & 0x20;
	twobpp = flags & 0x80;

	for(v = 0; v < 8; v++) {
		Uint16 c =  sprite[v] | (twobpp ? sprite[v + 8] : 0) << 8;
		for(h = 7; h >= 0; --h, c >>= 1) {
			Uint8 ch = (c & 1) | ((c >> 7) & 2);
			if(opaque || ch)
				screen_write(scr, layer,
					x + (flipx ? 7 - h : h),
					y + (flipy ? 7 - v : v),
					blending[ch][color]);
		}
	}
}

void
screen_palette(Screen *scr, Uint8 *addr)
{
	int i;
	int shift;

	for(i = 0, shift = 4; i < 4; i++, shift ^= 4) {
		Uint8 r, g, b;

		/* extract the colour values from r, g, b shorts */
		r = (addr[0 + i / 2] >> shift) & 0xf; r |= r << 4;
		g = (addr[2 + i / 2] >> shift) & 0xf; g |= g << 4;
		b = (addr[4 + i / 2] >> shift) & 0xf; b |= b << 4;

		r = r >> 3; r &= 0x1f;
		g = g >> 2; g &= 0x3f;
		b = b >> 3; b &= 0x1f;

		/* store into the palette */
		scr->palette[i] = (r << 11 | g << 5 | b);
	}

	scr->fg.changed = 1;
	scr->bg.changed = 1;
}

Uint8
screen_dei(Device *d, Uint8 port)
{
	switch(port) {
	case 0x2: return uxn_screen.width >> 8;
	case 0x3: return uxn_screen.width;
	case 0x4: return uxn_screen.height >> 8;
	case 0x5: return uxn_screen.height;
	}

	return d->dat[port];
}

void
screen_deo(Device *d, Uint8 port)
{
	Uint8 n, i;
	Uint8 twobpp;
	Uint16 x, y, dx, dy;
	Uint16 addr;

	Layer *layer;

	switch(port) {
	case 0xe:
		/* where */
		DEVPEEK16(x, 0x8);
		DEVPEEK16(y, 0xa);

		/* which layer */
		if(d->dat[0xe] & 0x40)
			layer = &uxn_screen.fg;
		else
			layer = &uxn_screen.bg;

		/* put and auto-increment if enabled */
		screen_write(&uxn_screen, layer, x, y, d->dat[0xe] & 0x3);
		if(d->dat[0x6] & 0x1)
			DEVPOKE16(0x8, x + 1);
		if(d->dat[0x6] & 0x2)
			DEVPOKE16(0xa, y + 1);

		break;
	case 0xf:
		/* where */
		DEVPEEK16(x, 0x8);
		DEVPEEK16(y, 0xa);
		DEVPEEK16(addr, 0xc);

		/* which layer */
		if(d->dat[0xf] & 0x40)
			layer = &uxn_screen.fg;
		else
			layer = &uxn_screen.bg;

		/* put */
		n = d->dat[0x6] >> 4;
		dx = (d->dat[0x6] & 0x1) << 3;
		dy = (d->dat[0x6] & 0x2) << 2;
		twobpp = !!(d->dat[0xf] & 0x80);
		if(addr > 0x10000 - ((n + 1) << (3 + twobpp)))
			return;

		for(i = 0; i <= n; i++) {
			screen_blit(&uxn_screen, layer,
				x + dy * i,
				y + dx * i,
				&d->u->ram[addr],
				d->dat[0xf] & 0xf,
				d->dat[0xf]);
			addr += (d->dat[0x6] & 0x4) << (1 + twobpp);
		}

		/* auto increment */
		DEVPOKE16(0xc, addr);
		DEVPOKE16(0x8, x + dx);
		DEVPOKE16(0xa, y + dy);
		break;
	}
}

